//
// file: primitive3D_searchLight.fx
// version: v1.5
// author: Ren712
//

//--------------------------------------------------------------------------------------
// Settings
//--------------------------------------------------------------------------------------
float3 sElementPosition = float3(0, 0, 0);
float3 sElementRotation = float3(0, 0, 0);
float4 sElementColor = float4(0, 0, 0, 0);
float sLightPhi = 0.5; // Phi is the outer cone angle
float sLightTheta = 0.3; // Theta is the inner cone angle
float sLightFalloff = 1; // light intensity attenuation between the phi and theta areas
float sLightAttenuation = 10; // light attenuation (max radius)
float sLightAttenuationPower = 2; // light attenuation power

float sCoronaRescale = 1;
float sCoronaDepthBias = 0.2;
float sCoronaScaleSpread = 0.85;

float2 gDistFade = float2(250, 150);

bool sFlattenTexture = true;
bool fZEnable = true;
bool fZWriteEnable = true;
int fCullMode = 1;

//--------------------------------------------------------------------------------------
// Variables set by MTA
//--------------------------------------------------------------------------------------
float4x4 gProjection : PROJECTION;
float4x4 gView : VIEW;
float4x4 gViewInverse : VIEWINVERSE;
float3 gCameraPosition : CAMERAPOSITION;

static const float PI = 3.14159265f;
int CUSTOMFLAGS < string skipUnusedParameters = "yes"; >;

//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------
struct VSInput
{
    float3 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float4 Diffuse : COLOR0;
};

struct PSInput
{
    float4 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float3 WorldPos : TEXCOORD1;
    float3 LightDir : TEXCOORD2;
    float DistFade : TEXCOORD3;
    float4 Diffuse : COLOR0;
};

//--------------------------------------------------------------------------------------
// Create world matrix with world position and ZXY rotation
//--------------------------------------------------------------------------------------
float4x4 createWorldMatrix(float3 pos, float3 rot)
{
    float4x4 eleMatrix = {
        float4(cos(rot.z) * cos(rot.y) - sin(rot.z) * sin(rot.x) * sin(rot.y), 
                cos(rot.y) * sin(rot.z) + cos(rot.z) * sin(rot.x) * sin(rot.y), -cos(rot.x) * sin(rot.y), 0),
        float4(-cos(rot.x) * sin(rot.z), cos(rot.z) * cos(rot.x), sin(rot.x), 0),
        float4(cos(rot.z) * sin(rot.y) + cos(rot.y) * sin(rot.z) * sin(rot.x), sin(rot.z) * sin(rot.y) - 
                cos(rot.z) * cos(rot.y) * sin(rot.x), cos(rot.x) * cos(rot.y), 0),
        float4(pos.x,pos.y,pos.z, 1),
    };
    return eleMatrix;
}

//--------------------------------------------------------------------------------------
// Return a rotation matrix (rotate by Y)
//--------------------------------------------------------------------------------------
float4x4 makeYRotation( float angleInRadians) 
{
  float c = cos(angleInRadians);
  float s = sin(angleInRadians);

  return float4x4(
    c, 0, -s, 0,
    0, 1, 0, 0,
    s, 0, c, 0,
    0, 0, 0, 1
  );
};

//--------------------------------------------------------------------------------------
// Vertex Shader 
//--------------------------------------------------------------------------------------
PSInput VertexShaderFunction(VSInput VS)
{	
    PSInput PS = (PSInput)0;
	
    // set proper position and scale of the quad
    VS.Position.xy = VS.TexCoord.xy;
    VS.Position.xy -= float2(0.5, 0.5);
    VS.Position.xy = - VS.Position.xy;
    VS.Position.z = 0; 
	
    VS.TexCoord.y = 1 - VS.TexCoord.y;
	
    // deform the quad shape into a triangle 
    if (VS.Position.x > 0) VS.Position.x += VS.Position.y;
        else VS.Position.x -= VS.Position.y;
    VS.Position.x *= 1 - 0.5;
    VS.Position.y += 0.5;

     // get texCoords from shape
    VS.TexCoord.xy = 1 - (0.5 - VS.Position.xy);
    VS.TexCoord = 1 - VS.TexCoord;
    VS.TexCoord = float2(1 - VS.TexCoord.x, VS.TexCoord.y + 0.5);
    VS.TexCoord.x = ((VS.TexCoord.x - 0.5) * VS.TexCoord.y) + 0.5;
	
    // get size
    float2 elementSize = float2(2 * tan(sLightPhi) * sLightAttenuation, sLightAttenuation);
	
    // move lower vertices a bit
    VS.Position.z -= elementSize.y * ((VS.Position.y) * tan(sLightTheta * 0.5));   

    // rescale
    VS.Position.xy *= elementSize;

    // create WorldMatrix for the quad
    float4x4 sWorld = createWorldMatrix(sElementPosition, sElementRotation);
    PS.LightDir = sWorld[1].xyz;
	
    // get element's view direction and distance from camera
    float3 viewDirection = (gCameraPosition - sElementPosition);
    float viewDistance = length(viewDirection);
	
    // is camera at the left/right side of the plane
    float planeYDist = dot(sWorld[0].xyz, viewDirection);
    float angYDir = planeYDist < 0 ?  1 : -1;
	
    // get angle between left plane and view
    float fwDist = dot(sWorld[1].xyz, viewDirection); 
    float3 sLinePosition = sElementPosition + sWorld[1].xyz * fwDist; 
    float angleY = acos(dot(normalize(sLinePosition - gCameraPosition), sWorld[2].xyz));
    angleY = angYDir * (angleY - PI);
	
    // rotate the material matrix to recreate cylindrical billboard effect
    sWorld = mul(makeYRotation(angleY), sWorld);
	
    // calculate screen position of the vertex
    float4 worldPos = mul(float4(VS.Position, 1), sWorld);
    float4 vPos = mul(worldPos, gView); 
    PS.Position = mul(vPos, gProjection);

    // get distance from front plane
    float planeZDist = dot(sWorld[1].xyz, viewDirection);
	
    // angle between front plane and view
    float angleZ = acos(dot(normalize(sElementPosition - gCameraPosition), sWorld[1].xyz));
    angleZ = planeZDist > 0 ? (PI - angleZ) : angleZ;

    float lightPhi = sLightPhi;
    float lightTheta = sLightTheta;

    // obscure the quad at a lower angle when vieved from back
    float distFactor = pow(saturate((tan(sLightPhi) * sLightAttenuation) / viewDistance), 2);
    distFactor = max(distFactor, 0.15);
    lightPhi = planeZDist > 0 ? lightPhi : lightPhi * distFactor;
    lightTheta = planeZDist > 0 ? lightTheta : lightTheta * distFactor;

    float impactZ = pow( smoothstep(lightPhi, lightTheta, angleZ), sLightFalloff);
    if (angleZ > lightPhi) impactZ = 0;
    impactZ = 1 - impactZ;
	
    // pass world position to pixel shader
    PS.WorldPos = worldPos.xyz;
	
    // get clip values
    float nearClip = - gProjection[3][2] / gProjection[2][2];
    float farClip = (gProjection[3][2] / (1 - gProjection[2][2]));	
	
    // fade corona
    float DistFromCam = distance(gCameraPosition, sElementPosition.xyz);
    float elementMedSize = (elementSize.x + elementSize.y) / 2;
    float2 DistFade = float2(min(gDistFade.x, farClip - elementMedSize / 2), min(gDistFade.y, farClip - elementMedSize /2 - (gDistFade.x - gDistFade.y)));
    PS.DistFade = saturate((DistFromCam - DistFade.x)/(DistFade.y - DistFade.x));
	
    // pass texCoords and vertex color to PS
    PS.TexCoord = VS.TexCoord;
    PS.Diffuse = VS.Diffuse * sElementColor;
    PS.Diffuse.a *= impactZ;	
	
    return PS;
}

PSInput VertexShaderFunction2(VSInput VS)
{
    PSInput PS = (PSInput)0;
	
    // set proper position and scale of the quad
    VS.Position.xy = VS.TexCoord.xy;
    VS.Position.xy -= float2(0.5, 0.5);
    VS.Position.xy = - VS.Position.xy;
    VS.Position.z = 0; 
	
	VS.TexCoord.y = 1 - VS.TexCoord.y;
	
    // scale the quad
    VS.Position.xy *= tan(sLightPhi) * sLightAttenuation * sCoronaRescale;
    VS.TexCoord.x = 1 - VS.TexCoord.x;

    // create WorldMatrix for the quad
    float4x4 sWorld = createWorldMatrix(sElementPosition, sElementRotation);
	
    // move the billboard a bit forward
    sWorld[3].xyz += sWorld[1].xyz * sCoronaDepthBias;
	
    // get distance from front plane
    float planeZDist = dot(sWorld[1].xyz, gCameraPosition - sElementPosition);
	
    // angle between front plane and view
    float angleZ = acos(dot(normalize(sElementPosition - gCameraPosition), sWorld[1].xyz));
	
    // deal with front/back case	
    angleZ = planeZDist > 0 ? (PI - angleZ) : PI / 2;
	
    float impactZ = pow( smoothstep(sLightPhi, sLightTheta, angleZ ), sLightFalloff);
    if (angleZ > sLightPhi) impactZ = 0;
	
    // scale the quad based on impactZ
    VS.Position.xy *= (1 - sCoronaScaleSpread) + impactZ * sCoronaScaleSpread;

    // calculate screen position of the vertex
    float4x4 sWorldView = mul(sWorld, gView);
    float3 positionVS = VS.Position.xyz + sWorldView[3].xyz;
    PS.Position = mul(float4(positionVS, 1), gProjection);
	
    // get quad size
    float2 elementSize = float2(2 * tan(sLightPhi) * sLightAttenuation, sLightAttenuation);

    // get clip values
    float nearClip = - gProjection[3][2] / gProjection[2][2];
    float farClip = (gProjection[3][2] / (1 - gProjection[2][2]));	
	
    // fade corona
    float DistFromCam = distance(gCameraPosition, sElementPosition.xyz);
    float elementMedSize = (elementSize.x + elementSize.y) / 2;
    float2 DistFade = float2(min(gDistFade.x, farClip - elementMedSize / 2), min(gDistFade.y, farClip - elementMedSize /2 - (gDistFade.x - gDistFade.y)));
    PS.DistFade = saturate((DistFromCam - DistFade.x)/(DistFade.y - DistFade.x));
	
    // pass texCoords and vertex color to PS
    PS.TexCoord = VS.TexCoord;
    PS.Diffuse = VS.Diffuse * sElementColor;

    return PS;
}

//--------------------------------------------------------------------------------------
// Pixel shaders 
//--------------------------------------------------------------------------------------
float4 PixelShaderFunction(PSInput PS) : COLOR0
{
    // compute the distance attenuation factor
    float fDistance = distance(sElementPosition, PS.WorldPos);
	
    // calculate linear attenuation (for the noisy cone)
    float fadeAmount = pow(saturate(fDistance / 1), 0.9);

    // compute the direction to the light
    float3 vLight = normalize(sElementPosition - PS.WorldPos);
	
    // compute the attenuation
    float fAttenuation = 1 - saturate(fDistance / sLightAttenuation);
    fAttenuation = pow(fAttenuation, sLightAttenuationPower);

    // determine the angle between the current sample
    // and the light's direction:
    float angle = acos(dot(-vLight, normalize(PS.LightDir)));
	
    // compute the spot attenuation factor
    float fSpotAtten = 0.0f;
    if ( angle > sLightPhi ) fSpotAtten = 0.0f;
    else if ( angle < sLightTheta) fSpotAtten = 1.0f;
    else fSpotAtten = pow( smoothstep(sLightPhi, sLightTheta, angle ), sLightFalloff);
	
    // ..if it's going to be a spotlight	 
    fAttenuation *= fSpotAtten * fadeAmount;
	
    // get vertex color
    float4 finalColor = PS.Diffuse;

    // multiply by attenuation and distFade
    finalColor.a *= saturate(fAttenuation);
    finalColor.a *= saturate(PS.DistFade);

    return saturate(finalColor);
}

float4 PixelShaderFunction2(PSInput PS) : COLOR0
{
    // compute the direction to the light
    float zDistance = tan(sLightPhi) * sLightAttenuation;
    float3 vLight = normalize(float3((PS.TexCoord.xy - 0.5) * zDistance, sLightAttenuation / 3));

    // determine the angle between the current sample
    // and the light's direction:
    float angle = acos(dot(-vLight, float3(0,0,-1)));
	
    // compute the spot attenuation factor
    float fSpotAtten = 0.0f;
    if ( angle > sLightPhi ) fSpotAtten = 0.0f;
    else if ( angle < sLightTheta) fSpotAtten = 1.0f;
    else fSpotAtten = pow( smoothstep(sLightPhi, sLightTheta, angle ), sLightFalloff);
	
    // get vertex color
    float4 finalColor = PS.Diffuse;

    // multiply by attenuation and distFade
    finalColor.a *= saturate(fSpotAtten);
    finalColor.a *= saturate(PS.DistFade);
    finalColor.a *= 0.8;

    return saturate(finalColor);
}

//--------------------------------------------------------------------------------------
// Techniques
//--------------------------------------------------------------------------------------
technique dxDrawPrimitive3Dfx_searchLight
{
  pass P0
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = fCullMode;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = false;
    VertexShader = compile vs_2_0 VertexShaderFunction();
    PixelShader  = compile ps_2_0 PixelShaderFunction();
  }
  
  pass P1
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = fCullMode;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = false;
    VertexShader = compile vs_2_0 VertexShaderFunction2();
    PixelShader  = compile ps_2_0 PixelShaderFunction2();
  }
} 

// Fallback
technique fallback
{
  pass P0
  {
    // Just draw normally
  }
}
